########################################################################
# ccp4idb
#
# Python module providing a class library for emulating the CCP4i
# def file based database.
#
# Version 0.0.2
#
########################################################################


#######################################################################
# Import modules that this module depends on
#######################################################################
import os
import re
import time

#######################################################################
# Class definitions
#######################################################################

# ccp4idb
#
# A class for handling CCP4i def file based databases
#
class ccp4idb:
    """CCP4i def file project database object"""

    def __init__(self):
        self.__project = ""
        self.__directory = ""
        self.__database = {}
        self.__loaded = False
        self.__modified = False

    def __repr__(self):
        return self.__project

    def create(self,project,directory):
        # Make a new database object
        dbdir = GetDbDir(directory)
        dbfile = GetDbFile(directory)
        if DirExists(dbdir) or FileExists(dbfile):
            print "Error: directory "+directory+" is already a project dir"
            return False
        if not DirExists(directory):
            print "Making directory "+directory
            MakeDir(directory)
        print "Making database subdirectory "+dbdir
        MakeDir(dbdir)
        print "Making database file "+dbfile
        MakeDbFile(dbfile)
        return self.open(project,directory)

    def open(self,project,directory):
        # Open an existing database object
        dbfile = GetDbFile(directory)
        if not FileExists(dbfile):
            print "Error: database file "+dbfile+" does not exist"
            return False
        self.__project = project
        self.__directory = directory
        self.__dblock = dblockfile(project,directory)
        if self.__dblock.islocked():
            print "Database file is locked by another process"
            return False
        self.__dblock.lock()
        self.__database = read_def_file(dbfile)
        self.__loaded = True
        return True

    def close(self):
        # Close an open database
        if not self.__loaded:
            print "Database for "+self.__project+" was never loaded"
            return False
        self.save()
        self.__dblock.unlock()
        self.__database = {}
        self.__loaded = False
        return True

    def selectjobs(self,pattern):
        print "selectjobs: not implemented"
        return

    def describe(self,joblist,itemlist,formatlist):
        print "describe: not implemented"
        return

    def njobs(self):
        # Return the value of NJOBS
        if not self.__loaded:
            return -1
        return int(self.__database["NJOBS"])

    def newjob(self):
        # Create a new record in the database
        if not self.__loaded:
            return -1
        n = self.njobs() + 1
        newrecord = MakeDbRecord(n)
        for item in newrecord.keys():
            self.__database[item] = newrecord[item]
        self.__database["NJOBS"] = n
        self.__modified = True
        return n

    def deletejob(self,job):
        # Remove a record from the database
        n = self.njobs()
        if job > n:
            return False
        ndel = 0
        for item in ListDbItems():
            key = item+","+str(job)
            if self.__database.has_key(key):
                del(self.__database[key])
                ndel = ndel + 1
        if ndel == 0:
            # Nothing was deleted
            return False
        if job == n:
            self.__database["NJOBS"] = n - 1
        self.__modified = True
        return True

    def itemexists(self,job,item):
        # Confirm whether or not an item exists for a job
        if not self.__loaded:
            return False
        if job > self.njobs():
            return False
        key = str(item)+","+str(job)
        return self.__database.has_key(key)

    def setdata(self,job,item,value):
        # Set the value of an item for a job
        if not self.itemexists(job,item):
            return False
        key = str(item)+","+str(job)
        self.__database[key] = value
        self.__modified = True
        return True

    def getdata(self,job,item):
        # Retrieve the value of an item
        if not self.itemexists(job,item):
            return ""
        key = str(item)+","+str(job)
        return self.__database[key]

    def describejob(self,job):
        # Placeholder: print out the values for items in a job
        nitems = 0
        description = ""
        for item in ListDbItems():
            if self.itemexists(job,item):
                description = description + " " + \
                              item + ":" + \
                              str(self.getdata(job,item))
                nitems = nitems + 1
        if nitems > 0:
            print "Description:", description
        else:
            print "No items set in this record"
        return

    def save(self):
        # Save the database contents to file
        dbfile = GetDbFile(self.__directory)
        write_def_file(dbfile,self.__database)
        self.__modified = False
        return True
        
# ccp4iarray
#
# A class for mimicking CCP4i-style associative arrays
# used for holding def file data
class ccp4iarray:
    """CCP4i parameter array object"""

    def __init__(self):
        print "ccp4iarray: class not implemented"

# dblockfile
#
# A class for handling database lockfiles
#
class dblockfile:
    """CCP4i database lock file object"""

    def __init__(self,project,directory):
        self.__project = project
        self.__directory = directory
        self.__lockfile = os.path.join(GetDbDir(directory),"database.LOCK")

    def filename(self):
        # Return the lockfile name (including the path)
        return self.__lockfile

    def islocked(self):
        # Check whether the database is locked
        return os.path.isfile(self.filename())

    def lock(self):
        # Lock the database directory
        if self.islocked():
            print "Error database is already locked"
            return 0
        # Create a lockfile
        try:
            f = open(self.filename(),"w")
        except:
            # Catch all errors
            print "Unable to open lockfile for writing"
            # Pass the error up to the calling application
            raise
        # Write the contents of the lockfile
        # From dbcore.tcl, lockfile conists of the following lines:
        # Lock file for <project>
        # Created by <uid>
        # Date <date>
        # Owned by pid <pid>
        f.write("Lock file for "+self.__project+"\n")
        f.write("Created by "+str(GetUserId())+"\n")
        f.write("Date "+str(GetDate())+"\n")
        f.write("Owned by pid "+str(GetPid())+"\n")
        f.close()
        return 1
        
    def unlock(self):
        # Unlock the database directory by removing the lockfile
        if not self.islocked():
            print "Error database is not currently locked"
            return 0
        # Delete the lockfile
        os.remove(self.filename())
        return 1

#######################################################################
# Supporting functions
#######################################################################

# Make a directory
def MakeDir(directory):
    return os.mkdir(directory)

# Check if a directory exists
def DirExists(directory):
    return os.path.isdir(directory)

# Check if a file exists
def FileExists(filen):
    return os.path.isfile(filen)

# Return the user id
def GetUserId():
    return os.environ.get("USER")

# Return the current date and time
def GetDate():
    return time.strftime("%d %b %Y  %H:%M:%S")

# Return the current process id
def GetPid():
    print "GetPid: FIXME, returning dummy value"
    return 1

# Return the CCP4 database directory given the project
# directory
def GetDbDir(projectdir):
    return os.path.join(projectdir,"CCP4_DATABASE")

# Return the CCP4 database.def file name given the
# project directory
def GetDbFile(projectdir):
    return os.path.join(GetDbDir(projectdir),"database.def")

# Make a new CCP4i database.def file
# This consists of the number of jobs plus a dummy record zero
def MakeDbFile(dbfile):
    dict = {}
    dict["NJOBS"] = 0
    record = MakeDbRecord(0)
    for item in record.keys():
        dict[item] = record[item]
    print dict
    return write_def_file(dbfile,dict)

# Make a database record
def MakeDbRecord(job_id):
    n = str(job_id)
    record = {}
    for item in ListDbItems():
        key = item+","+n
        record[key] = ""
    return record

# List the items in a database record
def ListDbItems():
    return ["TITLE", \
            "TASKNAME", \
            "DATE", \
            "STATUS", \
            "INPUT_FILES", \
            "INPUT_FILES_DIR", \
            "OUTPUT_FILES", \
            "OUTPUT_FILES_DIR", \
            "LOGFILE"]

# Read a CCP4i .def file and return a Python dictionary
def read_def_file(filename):
    """read_def_file(string)"""
    try:
        f = open(str(filename),"r")
    except:
        # Catch all errors
        print "read_def_file: error opening file \"%s\"" % str(filename)
        # Pass the error on to the calling subprogram
        raise
    # The def file consists of key-value pairs
    # Create a dictionary object to store the contents in
    contents = {}
    
    # Regular expressions...
    # ...for CCP4i comment lines
    comment = re.compile(r'^\s*#CCP4I')
    # ...for comment lines to be ignored
    ignore = re.compile(r'^\s*#')
    
    # Read the file a line at a time
    for line in f:
        # Remove leading spaces and trailing newline
        line = line.lstrip()
        line = line.rstrip("\n")
        # Ignore CCP4i comments (for now)
        if comment.match(line):
            ##print ">>> CCP4I comment line (ignored)"
            continue
        # Ignore comment lines (other lines starting with # symbol)
        if ignore.match(line):
            ##print ">>> Comment line (ignored)"
            continue
        # Ignore empty lines
        if len(line) < 1:
            ##print ">>> Empty line (ignored)"
            continue
        # All we are left with now are real lines
        fields = line.split(None,1)
        if len(fields) == 2:
            key = fields[0]
            value = fields[1]
            contents[key] = value
            ##print "Key = ",key," Value = ",value
    # Close the file
    f.close()
    # Return the dictionary
    return contents

# Write a CCP4i def file based on the contents
# of a dictionary
def write_def_file(filename,dict):
    """write_def_file(string)"""
    # Open a file for writing
    try:
        f = open(str(filename),"w")
    except:
        # Catch all errors
        print "write_def_file: error opening file \"%s\"" % str(filename)
        # Pass the error on to the calling subprogram
        raise
    # Write the header
    print "FIXME: write_def_file, header has no content"
    f.write("#CCP4I VERSION\n")
    f.write("#CCP4I SCRIPT DEF\n")
    f.write("#CCP4I DATE\n")
    f.write("#CCP4I USER\n")
    f.write("#CCP4I PROJECT\n")
    # The def file consists of key-value pairs
    # Write these one set per line to the file
    for key in dict.keys():
        f.write(str(key)+"\t"+str(dict[key])+"\n")
    # Close the file
    return 1
